import collections

import numpy as np


class Market:

    def __init__(self, a, a0, mu):
        self.a = np.array(a)
        self.a0 = a0
        self.mu = mu

    def demand(self, prices):
        x = np.exp((self.a - np.array(prices)) / self.mu)
        return x / (x.sum(-1, keepdims=True) + np.exp(self.a0 / self.mu))

    #def demand_table(self, pranges):
    #    ptable = np.array([[[p1, p2] for p2 in pranges[1]] for p1 in pranges[0]])
    #    return self.demand(ptable)
    
def paired_table(a, b):
    return np.stack(np.meshgrid(a, b), -1).transpose(1,0,2)

def monopoly(a):
    opt_sum = np.array(np.unravel_index(a.sum(2).argmax(), a[:,:,0].shape))
    return opt_sum

def nash(a):
    opt0 = a[:,:,0].argmax(0)
    opt1 = a[:,:,1].argmax(1)
    lst = []
    for i in (opt0[opt1] == np.arange(a.shape[0])).nonzero()[0]:
        j = opt1[i]
        lst.append([i, j])
    return np.array(lst)

class Env:

    def __init__(self, a, a0, mu, c, k, prices):
        self.n = len(a)
        self.market = Market(a, a0, mu)
        self.c = np.array(c)
        self.k = k
        self.prices = [np.array(p) for p in prices]

        self.price_table = paired_table(prices[0], prices[1])
        self.demand_table = self.market.demand(self.price_table)
        self.profit_table = (self.price_table - np.array(c)[np.newaxis, np.newaxis, :]) * self.demand_table

        self.reset()


    #@profile
    def step(self, actions):
        #print(actions)
        actions = sum(actions, ())
        self.history.append(actions)
        #p = np.array([self.prices[i][a] for i, a in enumerate(actions)])
        #d = self.market.demand(p)
        #pi = (p - self.c) * d
        #pi = self.profits[sum(actions, ())]
        pi = self.profit_table[actions]
        #o = tuple(np.array(self.history, dtype=np.int).flatten())
        o = sum(list(self.history), ())
        return [o, o], pi, False, {}

    def reset(self):
        self.history = collections.deque(maxlen=self.k)
        for _ in range(self.k):
            h = tuple([np.random.choice(len(self.prices[i])) for i in range(self.n)])
            self.history.append(h)
        o = sum(list(self.history), ())
        #o = tuple(np.array(self.history, dtype=np.int).flatten())
        return [o, o]

    def change_hist(self, seq_actions):
        for actions in seq_actions:
            self.history.append(actions)
        o = sum(list(self.history), ())
        #o = tuple(np.array(self.history, dtype=np.int).flatten())
        return [o, o]

    def render(self):
        print('--price history--')
        print([self.prices[idx] for idx in self.history])

        

if __name__ == '__main__':
    market = Market([1.5, 1.5], 0, 0.25)
    print(market.demand([1, 2]))

    env = Env([1.5, 1.5], 0, 0.25, c=[1, 1], k=3, prices=np.arange(1,16,2))
    env.render()
    print(env.step([1,3]))
    env.render()
    print(env.step([2,5]))
    env.render()
    print(env.step([0,1]))
    env.render()
    print(env.step([0,3]))
    env.render()

