import pickle
import random

import numpy as np

class ExponentialDecayEpsilonGreedy:

    def __init__(self, eps_start, eps_end, beta):
        self.eps_start = eps_start
        self.eps_end = eps_end
        self.beta = beta

    def __call__(self, t, q):
        eps = max(self.eps_start * np.exp(-self.beta * t), self.eps_end)
        if random.random() < eps:
            a = tuple([random.choice(list(range(d))) for d in q.shape])
        else:
            a =  np.unravel_index(np.argmax(q), q.shape)
        return a
    
class TableQLearning:

    def __init__(self, dim_state, dim_action, alpha, delta, explorer):
        self.qtable = np.zeros(dim_state + dim_action)
        self.alpha = alpha
        self.delta = delta
        self.dim_state = dim_state
        self.dim_action = dim_action
        self.explorer = explorer
        self.t = 0
        #self.eps = 1
        #self.expnegbeta = np.exp(-self.beta)
        self.training_mode()
    
    def freeze_q(self):
        self.update_q = False

    #@profile
    def act_greedy(self, obs):
        #print('g', obs, self.dim_action)
        qs = self.qtable[obs]
        #print(qs)
        return np.unravel_index(np.argmax(qs), self.dim_action)
   
    #@profile
    def act(self, obs):
        if not self.act_random:
            return self.act_greedy(obs)
        else:
            q = self.qtable[obs]
            a = self.explorer(self.t, q)
            self.last_obs = obs
            self.last_action = a
            self.t += 1
            return a

    #@profile
    def observe(self, obs, reward, done, reset):
        if self.update_q:
            tgt = reward + self.delta * self.qtable[obs].max()
            idx = self.last_obs + self.last_action
            self.qtable[idx] += self.alpha * (tgt - self.qtable[idx])

    def eval_mode(self):
        self.act_random = False
        self.update_q = False
        
    def training_mode(self):
        self.act_random  = True
        self.update_q = True
